from nometa.properties import Property
from lxml import etree


def _to_bool(val: str) -> bool:
    """
    Convert string to boolean.

    For these string values 'false','no','not','n','0' and '', it returns `False`. For any other value, it returns `True`.
    For any numeric representation >= '1.0', it returns `True`. Otherwise, `False`.
    """
    if not val: return False
    try:
        return not int(float(val))<1
    except ValueError:
        return not val.strip().lower() in ("false","no","not",'n','0','')


class BooleanProperty(Property):
    """
    Deal with boolean values. It can be used with any xml-tag data type, the assignment of value is determined as:

    For these string values 'false','no','not','n','0' and '', it puts `False`. For any other value, it puts `True`.
        
    For any numeric representation >= '1.0', it puts `True`. Otherwise, `False`.
    """
    def __init__(self, tname: str) -> None:
        """
        Constructor

        Args:
            tname: tag name in the XML file. Eg: dc:title 

        Raises:
            TypeError
        """
        super().__init__()
        if type(tname) != str:
            raise TypeError("`tname` must be string")
        
        self._tname=tname
        self.value=False
    
    @property
    def value(self) -> bool|None:
        """Get the value of this property"""
        return self._value
    
    @value.setter
    def value(self, val: bool|None) -> None:
        if val is None: pass
        elif type(val) != bool:
            raise TypeError("`value` must be boolean")
        
        self._value=val
        
    @staticmethod
    def from_element(tname: str, elem: etree._Element) -> Property:
        super(BooleanProperty, BooleanProperty).from_element(tname, elem)
        prop=BooleanProperty(tname)
        node=prop._find_node(elem)
        prop.value=_to_bool(node.text) if node is not None else None
        return prop
    
    def _set_node_value(self, node: etree._Element) -> None:
        super()._set_node_value(node)
        node.text="true" if self._value else "false" # type: ignore
        


from nometa.properties import Property
from lxml.etree import _Element



class TextProperty(Property):
    """
    Deal with string values. It reads and keeps all element value as string.
    """
    
    def __init__(self, tname: str) -> None:
        """
        Constructor

        Args:
            tname: tag name in the XML file. Eg: dc:title 

        Raises:
            TypeError
        """
        super().__init__()
        if type(tname) != str:
            raise TypeError("`tname` must be string")
        
        self._tname=tname
        self.value=None
    
    @property
    def value(self) -> str|None:
        """Get the string value of a property"""
        return self._value
    
    @value.setter
    def value(self, val: str|None) -> None:
        if val is None:
            self._value=val
            return
        elif type(val) != str:
            val=str(val)
        
        self._value=val.strip()
    
    @staticmethod
    def from_element(tname: str, elem: _Element) -> Property:
        super(TextProperty, TextProperty).from_element(tname, elem)
        prop=TextProperty(tname)
        node: _Element=prop._find_node(elem)
        prop.value=node.text if node is not None else None
        return prop
    
    def _set_node_value(self, node: _Element) -> None:
        super()._set_node_value(node)
        node.text=self._value # type: ignore



from nometa.properties.text import TextProperty
from nometa.properties.numeric import NumericProperty
from nometa.properties.datetime import DatetimeProperty
from nometa.properties.boolean import BooleanProperty
from datetime import datetime
from lxml import etree
from pytest import fixture, mark, raises

XML="""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cp:coreProperties
    xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties"
    xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/"
    xmlns:dcmitype="http://purl.org/dc/dcmitype/"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <dc:title></dc:title>
    <dc:subject></dc:subject>
    <dc:creator>Silverlayer Employee</dc:creator>
    <dc:description>mocked xml</dc:description>
    <cp:lastModifiedBy>Chuck Norris</cp:lastModifiedBy>
    <cp:revision>2</cp:revision>
    <dcterms:created xsi:type="dcterms:W3CDTF">2013-12-15T10:40:00Z</dcterms:created>
    <dcterms:modified xsi:type="dcterms:W3CDTF">2024-09-06T23:10:02Z</dcterms:modified>
    <cp:category></cp:category>
</cp:coreProperties>"""

@fixture
def xml_root() -> etree._Element:
    bxml=bytes(XML,"utf8")
    parser = etree.XMLParser(remove_blank_text=True, resolve_entities=False)
    parser.set_element_class_lookup(etree.ElementNamespaceClassLookup())
    return etree.fromstring(bxml, parser)


def test_wrong_initialization():
    check=[True,True,True,True,True]
    
    try: TextProperty(None); check[1]=False
    except TypeError: pass
    
    try: TextProperty(123); check[2]=False
    except TypeError: pass

    try: NumericProperty(None); check[3]=False
    except TypeError: pass

    try: DatetimeProperty(None); check[4]=False
    except TypeError: pass

    assert not False in check

def test_set_none_to_numeric_property():
    t=NumericProperty("cp:version")
    t.value=None
    assert t.value == None

def test_set_none_to_datetime_property():
    t=DatetimeProperty("dcterms:created")
    t.value=None
    assert t.value == None

def test_set_invalid_to_numeric_property():
    with raises(TypeError):
        t=NumericProperty("cp:version")
        t.value="23.4"

def test_set_invalid_to_datetime_property():
    with raises(TypeError):
        t=DatetimeProperty("dcterms:modified")
        t.value="2023-09-25T10:05:22"

def test_initial_integer_value(xml_root):
    t=NumericProperty.from_element("cp:revision", xml_root)
    assert t.value == int(2)

def test_initial_datetime_value(xml_root):
    created=datetime.strptime("2013-12-15T10:40:00Z","%Y-%m-%dT%H:%M:%SZ")
    modified=datetime.strptime("2024-09-06T23:10:02Z", "%Y-%m-%dT%H:%M:%SZ")
    c=DatetimeProperty.from_element("dcterms:created", xml_root)
    m=DatetimeProperty.from_element("dcterms:modified", xml_root)
    assert c.value == created and m.value == modified

def test_change_integer_value(xml_root):
    t=NumericProperty.from_element("cp:version", xml_root)
    t.value=1.34
    t.to_element(xml_root)
    u=NumericProperty.from_element("cp:version", xml_root)
    assert type(u.value) == float and u.value == 1.34

def test_change_datetime_value(xml_root):
    expected=datetime.strptime("2024-11-22T23:00:07Z","%Y-%m-%dT%H:%M:%SZ")
    t=DatetimeProperty.from_element("dcterms:modified", xml_root)
    t.value=expected
    t.to_element(xml_root)
    u=DatetimeProperty.from_element("dcterms:modified", xml_root)
    assert u.value == expected

def test_load_none_to_numeric_property(xml_root):
    t=NumericProperty.from_element("cp:category", xml_root)
    assert t.value == None

def test_load_none_to_datetime_property(xml_root):
    t=DatetimeProperty.from_element("cp:category", xml_root)
    assert t.value == None

def test_load_none_to_boolean_property(xml_root):
    t=BooleanProperty.from_element("cp:category", xml_root)
    assert t.value == False

def test_load_none_to_text_property(xml_root):
    t=TextProperty.from_element("cp:category", xml_root)
    assert t.value == None

@mark.parametrize("tname,expected",[
    ("dc:title",None), 
    ("dc:subject",None),
    ("cp:category",None),
    ("dc:creator","Silverlayer Employee"),
    ("dc:description","mocked xml"),
    ("cp:lastModifiedBy","Chuck Norris"),
    ("cp:revision",'2')
])
def test_initial_value(tname, expected, xml_root):
    t=TextProperty.from_element(tname, xml_root)
    if expected is None:
        assert t.value is expected
    else:
        assert t.value == expected


@mark.parametrize("tname",["dc:title", "dc:subject", "dc:creator", "dc:description", "cp:lastModifiedBy", "cp:category"])
def test_change_value(xml_root, tname):
    t=TextProperty.from_element(tname, xml_root)
    t.value=" 1984  "
    t.to_element(xml_root)
    u=TextProperty.from_element(tname, xml_root)
    assert u.value == "1984"

@mark.parametrize("tname",["dc:title", "dc:subject", "dc:creator", "dc:description", "cp:lastModifiedBy", "cp:category"])
def test_change_value_not_string(xml_root, tname):
    t=TextProperty.from_element(tname, xml_root)
    t.value=133
    t.to_element(xml_root)
    u=TextProperty.from_element(tname, xml_root)
    assert type(u.value) == str and u.value == "133"

def test_append_new_element(xml_root):
    expected="research, finance"
    new_prop=TextProperty("cp:keywords")
    new_prop.value=expected
    new_prop.to_element(xml_root)
    read_prop=TextProperty.from_element("cp:keywords",xml_root)
    assert read_prop.value == expected



from nometa.properties.boolean import _to_bool

def test_none_is_false():
    assert _to_bool(None) == False

def test_empty_string_is_false():
    assert _to_bool('') == False and _to_bool("   ") == False

def test_not_is_false():
    assert _to_bool("no") == False and _to_bool(" not") == False and _to_bool("n  ") == False

def test_0_is_false():
    assert _to_bool(0) == False and _to_bool('0') == False \
        and _to_bool(0.001) == False and _to_bool(0.999) == False \
        and _to_bool("0.001") == False and _to_bool("0.999") == False
    
def test_positive_number_is_true():
    """any number >= 1 is True"""
    assert _to_bool(1) == True and _to_bool('1') == True \
        and _to_bool('2.5') == True and _to_bool(73.3) == True


"""
This module contains the main XML sheets of metadata.
You can extend the App and Core classes in order to add new properties.
"""

from lxml import etree
from typing import cast
from abc import ABC, abstractmethod
from datetime import datetime
from nometa.properties.text import TextProperty
from nometa.properties.boolean import BooleanProperty
from nometa.properties.numeric import NumericProperty
from nometa.properties.datetime import DatetimeProperty


class Sheet(ABC):

    def __init__(self, raw: bytes) -> None:
        parser = etree.XMLParser(remove_blank_text=True, resolve_entities=False)
        parser.set_element_class_lookup(etree.ElementNamespaceClassLookup())
        self._xml_root=etree.fromstring(raw, parser)

    def pack(self) -> bytes:
        """
        Create the binary representation of `etree._Element`

        Returns:
            bytes: `etree._Element` as bytes
        """
        return etree.tostring(self._xml_root)
    
    @abstractmethod
    def to_element(self) -> None:
        """
        persist changes in this instance to the `etree._Element`
        """
        pass


class App(Sheet):

    def __init__(self, raw: bytes) -> None:
        """
        Create an instance of `App` representing the `docProps/app.xml` sheet.

        Args:
            raw (bytes): the `app.xml` sheet as binary
        """
        super().__init__(raw)
        node=self._xml_root
        self._template: TextProperty = cast(TextProperty, TextProperty.from_element("Template", node))
        self._total_time: NumericProperty = cast(NumericProperty, NumericProperty.from_element("TotalTime", node))
        self._application: TextProperty = cast(TextProperty, TextProperty.from_element("Application", node))
        self._scale_crop: BooleanProperty = cast(BooleanProperty, BooleanProperty.from_element("ScaleCrop", node))
        self._manager: TextProperty = cast(TextProperty, TextProperty.from_element("Manager", node))
        self._company: TextProperty = cast(TextProperty, TextProperty.from_element("Company", node))
        self._app_version: TextProperty = cast(TextProperty, TextProperty.from_element("AppVersion", node))

    def to_element(self) -> None:
        node=self._xml_root
        self._template.to_element(node)
        self._total_time.to_element(node)
        self._application.to_element(node)
        self._scale_crop.to_element(node)
        self._manager.to_element(node)
        self._company.to_element(node)
        self._app_version.to_element(node)

        return node
    
    @property
    def template(self) -> str|None:
        """
        The document's template. It's a read-only property.
        """
        return self._template.value
    
    @property
    def total_time(self) -> int|None:
        """The work time (measured in minutes)."""
        return cast(int, self._total_time.value)
    
    @total_time.setter
    def total_time(self, val: int) -> None:
        if type(val) != int:
            raise TypeError("``val`` must be an integer")
        
        self._total_time.value=val
    
    @property
    def application(self) -> str|None:
        """The description name of application"""
        return self._application.value
    
    @application.setter
    def application(self, val: str) -> None:
        self._application.value=val
    
    @property
    def scale_crop(self) -> bool|None:
        """
        Scale crop.
        
        This element indicates the display mode of the document thumbnail.
        Set this element to TRUE to enable scaling of the document thumbnail to the display. 
        Set this element to FALSE to enable cropping of the document thumbnail to show only sections that fits the display.
        """
        return self._scale_crop.value
    
    @scale_crop.setter
    def scale_crop(self, val: bool) -> None:
        self._scale_crop.value=val
    
    @property
    def manager(self) -> str|None:
        """Manager name"""
        return self._manager.value
    
    @manager.setter
    def manager(self, val: str) -> None:
        self._manager.value=val
    
    @property
    def company(self) -> str|None:
        """Company name"""
        return self._company.value
    
    @company.setter
    def company(self, val: str) -> None:
        self._company.value=val
    
    @property
    def app_version(self) -> str|None:
        """The application version. eg. 14.0000"""
        return self._app_version.value
    
    @app_version.setter
    def app_version(self, val: str) -> None:
        self._app_version.value=val


class Core(Sheet):

    def __init__(self, raw: bytes) -> None:
        """
        Create an instance of `Core` representing the `docProps/core.xml` sheet.

        Args:
            raw (bytes): the `core.xml` sheet as binary
        """
        super().__init__(raw)
        node=self._xml_root
        self._title: TextProperty = cast(TextProperty, TextProperty.from_element("dc:title", node))
        self._subject: TextProperty = cast(TextProperty, TextProperty.from_element("dc:subject", node))
        self._keyword: TextProperty = cast(TextProperty, TextProperty.from_element("cp:keywords", node))
        self._creator: TextProperty = cast(TextProperty, TextProperty.from_element("dc:creator", node))
        self._description: TextProperty = cast(TextProperty, TextProperty.from_element("dc:description", node))
        self._lastmdod: TextProperty = cast(TextProperty, TextProperty.from_element("cp:lastModifiedBy", node))
        self._category: TextProperty = cast(TextProperty, TextProperty.from_element("cp:category",node))
        self._revision: NumericProperty = cast(NumericProperty, NumericProperty.from_element("cp:revision",node))
        self._created: DatetimeProperty = cast(DatetimeProperty, DatetimeProperty.from_element("dcterms:created",node))
        self._modified: DatetimeProperty = cast(DatetimeProperty, DatetimeProperty.from_element("dcterms:modified",node))
        self._contentstat: TextProperty = cast(TextProperty, TextProperty.from_element("cp:contentStatus",node))
        self._version: TextProperty = cast(TextProperty, TextProperty.from_element("cp:version", node))
        self._identifier: TextProperty = cast(TextProperty, TextProperty.from_element("dc:identifier", node))
        self._last_printed: DatetimeProperty = cast(DatetimeProperty, DatetimeProperty.from_element("cp:lastPrinted", node))
        
    def to_element(self) -> None:
        node=self._xml_root
        self._title.to_element(node)
        self._subject.to_element(node)
        self._keyword.to_element(node)
        self._creator.to_element(node)
        self._description.to_element(node)
        self._lastmdod.to_element(node)
        self._category.to_element(node)
        self._revision.to_element(node)
        self._created.to_element(node)
        self._modified.to_element(node)
        self._contentstat.to_element(node)
        self._version.to_element(node)
        self._identifier.to_element(node)
        self._last_printed.to_element(node)

    @property
    def title(self) -> str|None:
        """The title"""
        return self._title.value
    
    @title.setter
    def title(self, val: str) -> None:
        self._title.value=val
    
    @property
    def subject(self) -> str|None:
        """The subject"""
        return self._subject.value
    
    @subject.setter
    def subject(self, val: str) -> None:
        self._subject.value=val
    
    @property
    def keywords(self) -> str|None:
        """
        The keywords.
        
        It is not compliance with CT_Keywords complex type specification.
        It's a simplified implementation that doesn't support multiple languages.
        """
        return self._keyword.value
    
    @keywords.setter
    def keywords(self, val: str) -> None:
        self._keyword.value=val

    @property
    def creator(self) -> str|None:
        """The creator/author"""
        return self._creator.value
    
    @creator.setter
    def creator(self, val: str) -> None:
        self._creator.value=val

    @property
    def description(self) -> str|None:
        """The description"""
        return self._description.value
    
    @description.setter
    def description(self, val: str) -> None:
        self._description.value=val

    @property
    def last_modified_by(self) -> str|None:
        """The last person that modified the document"""
        return self._lastmdod.value
    
    @last_modified_by.setter
    def last_modified_by(self, val: str) -> None:
        self._lastmdod.value=val

    @property
    def category(self) -> str|None:
        """The category"""
        return self._category.value
    
    @category.setter
    def category(self, val: str) -> None:
        self._category.value=val

    @property
    def revision(self) -> int|None:
        """The revision"""
        return cast(int, self._revision.value)
    
    @revision.setter
    def revision(self, val: int) -> None:
        if type(val) != int:
            raise TypeError("``val`` must be integer")
        
        self._revision.value=val

    @property
    def created(self) -> datetime|None:
        """When the document was created"""
        return self._created.value
    
    @created.setter
    def created(self, val: datetime) -> None:
        self._created.value=val

    @property
    def modified(self) -> datetime|None:
        """When the document was modified"""
        return self._modified.value
    
    @modified.setter
    def modified(self, val: datetime) -> None:
        self._modified.value=val

    @property
    def content_status(self) -> str|None:
        """
        The content status, its value should be 'Draft', 'Reviewed' or 'Final'.
        """
        return self._contentstat.value
    
    @content_status.setter
    def content_status(self, val: str) -> None:
        self._contentstat.value=val

    @property
    def version(self) -> str|None:
        """The version"""
        return self._version.value
    
    @version.setter
    def version(self, val: str) -> None:
        self._version.value=val
    
    @property
    def identifier(self) -> str|None:
        """The document's identifier. Only some documents apply it"""
        return self._identifier.value
    
    @identifier.setter
    def identifier(self, val: str) -> None:
        self._identifier.value=val

    @property
    def last_printed(self) -> datetime|None:
        """When the document was printed"""
        return self._last_printed.value
    
    @last_printed.setter
    def last_printed(self, val: datetime) -> None:
        self._last_printed.value=val


from nometa.sheet import Core
from datetime import datetime
from pytest import fixture, raises, mark

@fixture
def bxml() -> bytes:
    return bytes(
    """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    <cp:coreProperties
        xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties"
        xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/"
        xmlns:dcmitype="http://purl.org/dc/dcmitype/"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <dc:title>Insumos</dc:title>
        <dc:subject>Insumos hospitalares restantes</dc:subject>
        <dc:creator>Silverlayer Employee</dc:creator>
        <dc:description>mocked xml</dc:description>
        <cp:lastModifiedBy>Chuck Norris</cp:lastModifiedBy>
        <cp:revision>2</cp:revision>
        <dcterms:created xsi:type="dcterms:W3CDTF">2013-12-15T10:40:00Z</dcterms:created>
        <dcterms:modified xsi:type="dcterms:W3CDTF">2024-09-06T23:10:02Z</dcterms:modified>
        <cp:lastPrinted>2024-12-12T23:10:02Z</cp:lastPrinted>
        <cp:category>Controle</cp:category>
    </cp:coreProperties>""","utf8")

def test_init_core_sheet(bxml):
    osheet=Core(bxml)
    assert osheet is not None

@mark.parametrize("prop,expec",[
    ("title","Insumos"),
    ("subject","Insumos hospitalares restantes"),
    ("creator","Silverlayer Employee"),
    ("description","mocked xml"),
    ("last_modified_by","Chuck Norris"),
    ("revision",2),
    ("created",datetime.strptime("2013-12-15T10:40:00Z","%Y-%m-%dT%H:%M:%SZ")),
    ("modified",datetime.strptime("2024-09-06T23:10:02Z","%Y-%m-%dT%H:%M:%SZ")),
    ("last_printed", datetime.strptime("2024-12-12T23:10:02Z","%Y-%m-%dT%H:%M:%SZ")),
    ("category","Controle")
])
def test_read_props(bxml, prop, expec):
    osheet=Core(bxml)
    assert eval("osheet."+prop) == expec

def test_set_invalid_revision(bxml):
    osheet=Core(bxml)
    with raises(TypeError):
        osheet.revision="1.43.2"

def test_set_invalid_created(bxml):
    osheet=Core(bxml)
    with raises(TypeError):
        osheet.created="undefined"

def test_set_invalid_modified(bxml):
    osheet=Core(bxml)
    with raises(TypeError):
        osheet.modified="11/09/2001T13:01:00Z"

def test_change_props(bxml):
    osheet=Core(bxml)
    osheet.title="My wonderful world"
    osheet.subject="nothing"
    osheet.creator=None
    osheet.description=None
    osheet.last_modified_by="popeye"
    osheet.revision=77
    osheet.created=datetime.strptime("1889-05-29T20:55:43Z","%Y-%m-%dT%H:%M:%SZ")
    osheet.modified=datetime.strptime("2023-12-31T09:25:50Z","%Y-%m-%dT%H:%M:%SZ")
    osheet.category=None
    osheet.to_element()
    new_sheet=Core(osheet.pack())

    assert new_sheet.title == osheet.title and new_sheet.subject == osheet.subject and new_sheet.last_printed == osheet.last_printed \
    and new_sheet.creator == osheet.creator and new_sheet.description == osheet.description and new_sheet.revision == osheet.revision \
    and new_sheet.last_modified_by == osheet.last_modified_by and new_sheet.category == osheet.category and new_sheet.created == osheet.created


"""
NoMETA is a library for editing MS Office documents metadata,
it works with documents of Office >= 2007 and has been tested with docx, pptx, xlsx, vsdx and accdt files.    
"""

from typing import Type, IO
from nometa.sheet import App, Core
from zipfile import is_zipfile, ZipFile, ZIP_DEFLATED

__version__="0.1.1"

def _copy_archive(zin: ZipFile, zout: ZipFile) -> None:
    zout.comment=zin.comment
    for it in zin.infolist():
        if it.filename in ["docProps/app.xml", "docProps/core.xml"]: continue
        zout.writestr(it, zin.read(it))

class Document:
    """
    The Document class aggregates two sheets. One sheet represents the *docProps/app.xml* and the other one represents the *docProps/core.xml*
    """
    def __init__(self, file: str | IO[bytes], cls_core: Type[Core], cls_app: Type[App]) -> None:
        """
        Open the specified document to handle its metadata.

        Args:
            file (str | IO[bytes]): file path of document as string OR an in-memory buffer (`io.BytesIO`)
            cls_core (Type[Core]): type of `nometa.sheet.App` class or its subclasses
            cls_app (Type[App]): type of `nometa.sheet.Core` class or its subclasses

        Raises:
            TypeError
            ValueError
            ValueError
        """
        if not (issubclass(cls_core, Core) and issubclass(cls_app, App)):
            raise TypeError("``cls_app`` must be a subclass of ``sheet.App`` and ``cls_core`` must be a subclass of ``sheet.Core``")
        
        if not is_zipfile(file):
            raise ValueError("'%s' is not a MS Office document"%file)
        
        with ZipFile(file, 'r') as zf:
            try:
                self._core=cls_core(zf.read("docProps/core.xml"))
            except KeyError:
                raise ValueError("'%s' is not a MS Office document"%file)
            
            try:
                self._app=cls_app(zf.read("docProps/app.xml"))
            except KeyError:
                self._app=None
            
            self._file: str | IO[bytes]=file

    @property
    def app(self) -> App:
        """
        Get `App` sheet instance correponding to `docProps/app.xml` file.

        Raises:
            NotImplementedError: throws when there is no `App` instance.

        Returns:
            App: an instance of `nometa.sheet.App`
        """
        if self._app is None:
            raise NotImplementedError("This document doesn't have app.xml sheet")
        
        return self._app

    @property
    def core(self) -> Core:
        """
        Get `Core` sheet instance corresponding to `docProps/core.xml` file.

        Returns:
            Core: an instance of `nometa.sheet.Core`
        """
        return self._core
            
    def save(self, outfile: str|IO[bytes]) -> None:
        """
        Save the changes to the specified document in `outfile` parameter.

        Args:
            outfile (str | IO[bytes]): file path (as string) to document or in memory buffer (`io.BytesIO`)

        Raises:
            IOError: throws when input and output file path/buffer are the same
        """
        if self._file == outfile:
            raise IOError("Input and output documents cannot be the same")
        
        with ZipFile(self._file, 'r') as zr:
            with ZipFile(outfile, 'w', compression=ZIP_DEFLATED) as zw:
                _copy_archive(zr,zw)
                self._core.to_element()
                zw.writestr("docProps/core.xml",self._core.pack())
                if self._app is not None:
                    self._app.to_element()
                    zw.writestr("docProps/app.xml", self._app.pack())



__all__ = ["Document", "App", "Core", "__version__"]


from nometa.sheet import App
from pytest import fixture, raises, mark

@fixture
def bxml() -> bytes:
    return bytes("""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    <Properties xmlns="http://schemas.openxmlformats.org/officeDocument/2006/extended-properties"
        xmlns:vt="http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes">
        <Template>Normal.dotm</Template>
        <TotalTime>93</TotalTime>
        <Pages>1</Pages>
        <Words>7</Words>
        <Characters>41</Characters>
        <Application>Microsoft Macintosh Word</Application>
        <DocSecurity>0</DocSecurity>
        <Lines>1</Lines>
        <Paragraphs>1</Paragraphs>
        <ScaleCrop>false</ScaleCrop>
        <HeadingPairs>
            <vt:vector size="2" baseType="variant">
                <vt:variant>
                    <vt:lpstr>Title</vt:lpstr>
                </vt:variant>
                <vt:variant>
                    <vt:i4>1</vt:i4>
                </vt:variant>
            </vt:vector>
        </HeadingPairs>
        <TitlesOfParts>
            <vt:vector size="1" baseType="lpstr">
                <vt:lpstr></vt:lpstr>
            </vt:vector>
        </TitlesOfParts>
        <Manager>Josh Knuck</Manager>
        <Company>Silverlayer</Company>
        <LinksUpToDate>false</LinksUpToDate>
        <CharactersWithSpaces>47</CharactersWithSpaces>
        <SharedDoc>false</SharedDoc>
        <HyperlinkBase></HyperlinkBase>
        <HyperlinksChanged>false</HyperlinksChanged>
        <AppVersion>14.0000</AppVersion>
    </Properties>
    ""","utf8")

def test_init_app_sheet(bxml):
    osheet=App(bxml)
    assert osheet is not None

@mark.parametrize("prop,expec",[
    ("template","Normal.dotm"),
    ("total_time",93),
    ("application","Microsoft Macintosh Word"),
    ("scale_crop",False),
    ("manager","Josh Knuck"),
    ("company","Silverlayer"),
    ("app_version","14.0000")
])
def test_read_props(bxml, prop, expec):
    osheet=App(bxml)
    assert eval("osheet."+prop) == expec


def test_set_invalid_total_time(bxml):
    osheet=App(bxml)
    with raises(TypeError):
        osheet.total_time=3.1456795

def test_set_invalid_scale(bxml):
    osheet=App(bxml)
    with raises(TypeError):
        osheet.scale_crop="True"

def test_change_prop(bxml):
    osheet=App(bxml)
    osheet.manager="Michael Rudson"
    osheet.scale_crop=True
    osheet.total_time=240
    osheet.application="MS Office"
    osheet.company="Neo brand"
    osheet.app_version="12.000"
    osheet.to_element()
    new_sheet=App(osheet.pack())
    
    assert new_sheet.manager == osheet.manager and new_sheet.scale_crop == osheet.scale_crop \
    and new_sheet.total_time == osheet.total_time and new_sheet.application == osheet.application \
    and new_sheet.company == osheet.company and new_sheet.app_version == osheet.app_version


from nometa.properties import Property
from lxml.etree import _Element


class NumericProperty(Property):
    """
    Deal with numeric (integer or float) values. It must be used only in float or integer xml elements.
    """

    def __init__(self, tname: str) -> None:
        """
        Constructor

        Args:
            tname: tag name in the XML file. Eg: dc:title 

        Raises:
            TypeError
        """
        super().__init__()
        if type(tname) != str:
            raise TypeError("`tname` must be string")
        
        self._tname=tname
        self.value=None
    
    @property
    def value(self) -> int|float|None:
        """Get the value of this property"""
        return self._value
    
    @value.setter
    def value(self, val: int|float|None) -> None:
        if val is None: pass
        elif type(val) != int and type(val) != float:
            raise TypeError("type of `value` must be int or float")
        
        self._value=val
    
    @staticmethod
    def from_element(tname: str, elem: _Element) -> Property:
        super(NumericProperty, NumericProperty).from_element(tname, elem)
        prop=NumericProperty(tname)
        node=prop._find_node(elem)
        try:
            prop.value=int(node.text)
        except (TypeError, AttributeError):
            prop.value=None
        except ValueError:
            try:
                prop.value=float(node.text)
            except TypeError:
                prop.value=None
            except ValueError:
                raise ValueError("The tag value isn't numeric")
        
        return prop
    
    
    def _set_node_value(self, node: _Element) -> None:
        super()._set_node_value(node)
        node.text=str(self._value) # type: ignore
        


from nometa.properties import Property
from datetime import datetime
from lxml.etree import _Element



class DatetimeProperty(Property):
    """
    Deal with UTC-ISO formatted dates. It must be used only with ISO formatted dates, and always writes UTC time.
    """

    def __init__(self, tname: str) -> None:
        """
        Constructor

        Args:
            tname: tag name in the XML file. Eg: dc:title 

        Raises:
            TypeError
        """
        super().__init__()
        if type(tname) != str:
            raise TypeError("`tname` must be string")
        
        self._tname=tname
        self.value=None
    
    @property
    def value(self) -> datetime|None:
        """Get a `datetime` instance in UTC"""
        return self._value
    
    @value.setter
    def value(self, val: datetime|None) -> None:
        if val is None: pass
        elif not isinstance(val, datetime):
            raise TypeError("`value` must be a valid instance of `datetime`")
        
        self._value=val
    
    @staticmethod
    def from_element(tname: str, elem: _Element) -> Property:
        super(DatetimeProperty, DatetimeProperty).from_element(tname, elem)
        prop=DatetimeProperty(tname)
        node=prop._find_node(elem)
        prop.value=datetime.strptime(node.text,"%Y-%m-%dT%H:%M:%SZ") if (node is not None and node.text is not None) else None
        return prop
    
    def _set_node_value(self, node: _Element) -> None:
        super()._set_node_value(node)
        node.text=self._value.strftime("%Y-%m-%dT%H:%M:%SZ") # type: ignore
        


"""
This subpackage deal with XML tags of diverse data type.
It contains classes to handle tags whose data type are: boolean, datetime, numeric and text.
"""

from abc import ABC, abstractmethod
from lxml import etree
from typing import Any


class Property(ABC):
    """
    Abstract class for dealing with xml nodes/tags
    """
    def __init__(self) -> None:
        self._tname: str = ''
        self._value: Any = None

    @property
    def tname(self) -> str:
        """Get the official tag name of this property"""
        return self._tname

    @staticmethod
    @abstractmethod
    def from_element(tname: str, elem: etree._Element) -> "Property":
        """
        Read an `_Element` and returns the corresponding property object

        Args:
            tname (str): the official tag name in XML documents
            elem (etree._Element): a valid `etree._Element` instance

        Raises:
            TypeError

        Returns:
            Property: an instance of `Property` or its subclasses
        """
        if type(tname) != str:
            raise TypeError("`tname` must be string")
        
        if type(elem) != etree._Element:
            raise TypeError("`elem` must be an instance of `etree._Element`")

    
    def __append(self, elem: etree._Element) -> None:
        if ':' in self._tname:
            xmlns, tag=self._tname.split(':')
            new_node=elem.makeelement(etree.QName(elem.nsmap[xmlns], tag)) # type: ignore
        else:
            new_node=elem.makeelement(self._tname)  # type: ignore
        
        self._set_node_value(new_node)
        elem.append(new_node)

    def __update(self, elem: etree._Element) -> None:
        node=self._find_node(elem)
        self._set_node_value(node)

    def _find_node(self, elem: etree._Element) -> etree._Element:
        node: etree._Element = elem.find(self._tname, namespaces=elem.nsmap)
        if node is None:
            try:
                node=elem.find(self._tname)
            except SyntaxError:
                node=None

        return node

    @abstractmethod
    def _set_node_value(self, node: etree._Element) -> None:
        """
        Hook method to set the tag's value

        Args:
            node (etree._Element): the node corresponding this XML tag
        """
        pass

    def to_element(self, elem: etree._Element) -> None:
        """
        Write the value of this property to `elem` parameter

        Args:
            elem (etree._Element): the target xml tree node

        Raises:
            TypeError
        """
        if type(elem) != etree._Element:
            raise TypeError("`elem` must be an instance of `etree._Element`")
        
        node=elem.find(self._tname, namespaces=elem.nsmap)
        if node is not None:
            self.__update(elem)
        elif node is None and self._value is not None:
            self.__append(elem)
            



__all__ = ["Property"]


from nometa.properties.text import TextProperty
from typing import cast
from pytest import fixture
from nometa import *

@fixture
def bxml() -> bytes:
    return bytes(
    """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    <cp:coreProperties
        xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties"
        xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/"
        xmlns:dcmitype="http://purl.org/dc/dcmitype/"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <dc:title>Insumos</dc:title>
        <dc:subject>Insumos hospitalares restantes</dc:subject>
        <dc:creator>Silverlayer Employee</dc:creator>
        <dc:unreal>Anything</dc:unreal>
        <dc:description>mocked xml</dc:description>
        <cp:lastModifiedBy>Chuck Norris</cp:lastModifiedBy>
        <cp:revision>2</cp:revision>
        <dcterms:created xsi:type="dcterms:W3CDTF">2013-12-15T10:40:00Z</dcterms:created>
        <dcterms:modified xsi:type="dcterms:W3CDTF">2024-09-06T23:10:02Z</dcterms:modified>
        <cp:lastPrinted>2024-12-12T23:10:02Z</cp:lastPrinted>
        <cp:category>Controle</cp:category>
    </cp:coreProperties>""","utf8")


class UnrealCore(Core):
    def __init__(self, raw) -> None:
        super().__init__(raw)
        self._unreal=TextProperty.from_element("dc:unreal", self._xml_root)

    def to_element(self) -> None:
        super().to_element()
        self._unreal.to_element(self._xml_root)

    @property
    def unreal(self) -> str|None:
        return self._unreal.value
    
    @unreal.setter
    def unreal(self, val: str) -> None:
        self._unreal.value=val
    

def test_extended_sheet_with_new_property(bxml) -> None:
    prop=UnrealCore(bxml)
    assert prop.unreal == "Anything" and prop.category == "Controle"

def test_open_document_with_unreal_class() -> None:
    doc=Document("tests/resource/test.xlsx",UnrealCore,App)
    ucore=cast(UnrealCore, doc.core)
    assert ucore.unreal is None and ucore.modified.year == 2024

def test_save_doc_with_unreal_prop() -> None:
    doc=Document("tests/resource/test.xlsx",UnrealCore,App)
    ucore=cast(UnrealCore, doc.core)
    ucore.unreal="paradise"
    doc.save("tests/resource/test20.xlsx")


from nometa import *
from pathlib import Path
from datetime import datetime
from argparse import ArgumentParser, SUPPRESS, RawDescriptionHelpFormatter
import sys

_MESSAGE=f"""
***********************************************************************************************************************************
* © 2024 Silverlayer
* NoMETA v{__version__}
* 
* NoMETA edits MS Office documents metadata (Office>=2007). On success, it will give a copy of provided document with modfications.
* To clean the properties `creator|last_modified_by|last_printed|manager|company`, use the corresponding option with empty string.
* Tested on documents *.docx, *.xlsx, *.pptx, *.vsdx and *.accdt.
* 
* This software is licensed under the Apache License version 2.0 (http://www.apache.org/licenses/LICENSE-2.0)
***********************************************************************************************************************************
Eg.
    nometa --creator '' sample.docx --> clean property creator
    nometa --created 2024-12-09T13:09:23 --manager 'Josh Kool' sample.xlsx --> set properties `created` and `manager` at same time.
"""

def main() -> None:
    parser=ArgumentParser(
        prog="nometa",
        description=_MESSAGE,
        allow_abbrev=False,
        formatter_class=RawDescriptionHelpFormatter
    )

    dt_conv=lambda dts: datetime.strptime(dts,"%Y-%m-%dT%H:%M:%SZ")

    parser.add_argument("docpath", type=str, help="The document path")
    # properties of core.xml
    parser.add_argument("--creator", default=SUPPRESS, help="who has created the document")
    parser.add_argument("--last_modified_by", default=SUPPRESS, help="who has modified the document")
    parser.add_argument("--created", default=SUPPRESS, type=dt_conv, help="when the doc. has been created (ISO 8601 UTC)")
    parser.add_argument("--modified", default=SUPPRESS, type=dt_conv, help="when the doc. has been modified (ISO 8601 UTC)")
    parser.add_argument("--last_printed", default=SUPPRESS, help="when the doc. has been printed")

    # properties of app.xml
    parser.add_argument("--manager", default=SUPPRESS, help="manager's name")
    parser.add_argument("--company", default=SUPPRESS, help="company's name")
    args=parser.parse_args()
    dargs=vars(args)

    docfile=Path(dargs["docpath"]).expanduser()
    del dargs["docpath"]

    if not docfile.is_file():
        raise FileNotFoundError("file not found at '%s'"%docfile)

    doc=Document(str(docfile),Core,App)

    for prop,value in dargs.items():
        if prop in ["creator","last_modified_by"]:
            setattr(doc.core,prop,None if len(value)<=0 else value)
        elif prop in ["manager","company"]:
            setattr(doc.app,prop,None if len(value)<=0 else value)
        elif prop == "last_printed":
            setattr(doc.core,prop,datetime.fromisoformat(value) if len(value)>0 else None)
        else:
            setattr(doc.core,prop,value)

    outfile=docfile.name.replace(docfile.suffix,'')+"_copy"+docfile.suffix
    doc.save(outfile)

if __name__ == "__main__":
    rc: int = 1
    try:
        main()
        rc = 0
    except Exception as e:
        print("Err: %s"%e, file=sys.stderr)
    finally:
        sys.exit(rc)


from nometa.properties.text import TextProperty
from nometa.properties.numeric import NumericProperty
from nometa.properties.boolean import BooleanProperty
from lxml import etree
from pytest import fixture, mark, raises

XML=bytes("""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<Properties xmlns="http://schemas.openxmlformats.org/officeDocument/2006/extended-properties"
    xmlns:vt="http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes">
    <Template>Normal.dotm</Template>
    <TotalTime>93</TotalTime>
    <Pages>1</Pages>
    <Words>7</Words>
    <Characters>41</Characters>
    <Application>Microsoft Macintosh Word</Application>
    <DocSecurity>0</DocSecurity>
    <Lines>1</Lines>
    <Paragraphs>1</Paragraphs>
    <ScaleCrop>false</ScaleCrop>
    <HeadingPairs>
        <vt:vector size="2" baseType="variant">
            <vt:variant>
                <vt:lpstr>Title</vt:lpstr>
            </vt:variant>
            <vt:variant>
                <vt:i4>1</vt:i4>
            </vt:variant>
        </vt:vector>
    </HeadingPairs>
    <TitlesOfParts>
        <vt:vector size="1" baseType="lpstr">
            <vt:lpstr></vt:lpstr>
        </vt:vector>
    </TitlesOfParts>
    <Manager>Josh Knuck</Manager>
    <Company>Silverlayer</Company>
    <LinksUpToDate>false</LinksUpToDate>
    <CharactersWithSpaces>47</CharactersWithSpaces>
    <SharedDoc>false</SharedDoc>
    <HyperlinkBase></HyperlinkBase>
    <AppVersion>14.0000</AppVersion>
</Properties>
""","utf8")

@fixture
def xml_root() -> etree._Element:
    parser = etree.XMLParser(remove_blank_text=True, resolve_entities=False)
    parser.set_element_class_lookup(etree.ElementNamespaceClassLookup())
    return etree.fromstring(XML, parser)


@mark.parametrize("tname,expected",[
    ("Template", "Normal.dotm"),
    ("Application", "Microsoft Macintosh Word"),
    ("Manager","Josh Knuck"),
    ("Company", "Silverlayer"),
    ("AppVersion", "14.0000")
])
def test_init_text_value(tname,expected,xml_root):
    t=TextProperty.from_element(tname, xml_root)
    assert t.value == expected

def test_init_integer_value(xml_root):
    total=NumericProperty.from_element("TotalTime", xml_root)
    assert total.value == 93

def test_init_boolean_value(xml_root):
    scale=BooleanProperty.from_element("ScaleCrop", xml_root)
    assert scale.value == False

def test_change_scale_crop(xml_root):
    scale=BooleanProperty.from_element("ScaleCrop", xml_root)
    scale.value=True
    scale.to_element(xml_root)
    scale2=BooleanProperty.from_element("ScaleCrop",xml_root)
    assert scale2.value == True

def test_change_scale_crop_none(xml_root):
    scale=BooleanProperty.from_element("ScaleCrop", xml_root)
    scale.value=None
    scale.to_element(xml_root)
    scale2=BooleanProperty.from_element("ScaleCrop",xml_root)
    assert scale2.value == False

def test_set_invalid_to_boolean():
    tst=BooleanProperty("ScaleCrop")
    with raises(TypeError):
        tst.value="True"


from nometa import Document
from nometa.sheet import App, Core
from pytest import mark, raises
import io

RESOURCE_PATH="tests/resource/"

@mark.parametrize("file",["test.docx","test.xlsx","test.pptx","test.vsdx"])
def test_init_doc(file):
    doc=Document(RESOURCE_PATH+file,Core,App)
    assert doc.core.modified.year == 2024 and float(doc.app.app_version)>=14.0

def test_doc_without_app_sheet():
    doc=Document(RESOURCE_PATH+"test.accdt",Core,App)
    with raises(NotImplementedError):
        doc.app

def test_save_inplace_doc():
    doc=Document(RESOURCE_PATH+"test.docx",Core,App)
    doc.app.total_time=3000
    doc.core.creator="Papaléguas"
    with raises(IOError):
        doc.save(RESOURCE_PATH+"test.docx")

@mark.parametrize("src,dst",[
    ("test.docx","test2.docx"),
    ("test.xlsx","test2.xlsx"),
    ("test.pptx","test2.pptx"),
    ("test.vsdx","test2.vsdx"),
    ("test.accdt","test2.accdt")
])
def test_save_doc(src, dst):
    doc=Document(RESOURCE_PATH+src,Core,App)
    try:
        doc.app.total_time=127
        doc.app.company="Silverlayer"
        doc.app.manager=None
    except NotImplementedError: pass
    doc.core.creator = doc.core.last_modified_by = "Johnny Test"
    doc.save(RESOURCE_PATH+dst)

def test_doc_asbytes():
    buff=io.BytesIO()
    with open(RESOURCE_PATH+"test.docx", "rb") as fd:
        doc=Document(fd,Core,App)
        doc.app.total_time=25
        doc.save(buff)
    
    new_doc=Document(buff,Core,App)
    assert doc.app.total_time == new_doc.app.total_time and doc.core.creator == new_doc.core.creator

def test_read_write_same_buffer():
    with open(RESOURCE_PATH+"test.docx", "rb") as fd:
        doc=Document(fd,Core,App)
        doc.app.total_time=25
        with raises(IOError):
            doc.save(fd)


def test_filepath_to_bytes():
    buff=io.BytesIO()
    doc=Document(RESOURCE_PATH+"test.docx",Core,App)
    doc.app.total_time=25
    doc.save(buff)
    
    new_doc=Document(buff,Core,App)
    assert doc.app.total_time == new_doc.app.total_time and doc.core.creator == new_doc.core.creator
